/*
 * Playing-field leveller for Build
 * by Jonathon Fowler
 *
 * A note about this:
 * 1. There is some kind of problem somewhere in the functions below because
 *    compiling with __compat_h_macrodef__ disabled makes stupid things happen.
 * 2. The functions below, aside from the ones which aren't trivial, were never
 *    really intended to be used for anything except tracking anr removing ties
 *    to the standard C library from games. Using the Bxx versions of functions
 *    means we can redefine those names to link up with different runtime library
 *    names.
 */

#ifdef WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#if defined(__WATCOMC__)
# include <direct.h>
#elif defined(_MSC_VER)
# include <io.h>
#else
# include <dirent.h>
#endif

#include "compat.h"


#ifndef __compat_h_macrodef__

int Brand(void)
{
	return rand();
}

void *Bmalloc(bsize_t size)
{
	return malloc(size);
}

void Bfree(void *ptr)
{
	free(ptr);
}

int Bopen(const char *pathname, int flags, unsigned mode)
{
	int n=0,o=0;
	
	if (flags&BO_BINARY) n|=O_TEXT; else n|=O_BINARY;
	if ((flags&BO_RDWR)==BO_RDWR) n|=O_RDWR;
	else if ((flags&BO_RDWR)==BO_WRONLY) n|=O_WRONLY;
	else if ((flags&BO_RDWR)==BO_RDONLY) n|=O_RDONLY;
	if (flags&BO_APPEND) n|=O_APPEND;
	if (flags&BO_CREAT) n|=O_CREAT;
	if (flags&BO_TRUNC) n|=O_TRUNC;
	if (mode&BS_IREAD) o|=S_IREAD;
	if (mode&BS_IWRITE) o|=S_IWRITE;
	if (mode&BS_IEXEC) o|=S_IEXEC;
	
	return open(pathname,n,o);
}

int Bclose(int fd)
{
	return close(fd);
}

bssize_t Bwrite(int fd, const void *buf, bsize_t count)
{
	return write(fd,buf,count);
}

bssize_t Bread(int fd, void *buf, bsize_t count)
{
	return read(fd,buf,count);
}

int Blseek(int fildes, int offset, int whence)
{
	switch (whence) {
		case BSEEK_SET: whence=SEEK_SET; break;
		case BSEEK_CUR: whence=SEEK_CUR; break;
		case BSEEK_END: whence=SEEK_END; break;
	}
	return lseek(fildes,offset,whence);
}

BFILE *Bfopen(const char *path, const char *mode)
{
	return (BFILE*)fopen(path,mode);
}

int Bfclose(BFILE *stream)
{
	return fclose((FILE*)stream);
}

void Brewind(BFILE *stream)
{
	rewind((FILE*)stream);
}

int Bfgetc(BFILE *stream)
{
	return fgetc((FILE*)stream);
}

char *Bfgets(char *s, int size, BFILE *stream)
{
	return fgets(s,size,(FILE*)stream);
}

int Bfputc(int c, BFILE *stream)
{
	return fputc(c,(FILE*)stream);
}

int Bfputs(const char *s, BFILE *stream)
{
	return fputs(s,(FILE*)stream);
}

bsize_t Bfread(void *ptr, bsize_t size, bsize_t nmemb, BFILE *stream)
{
	return fread(ptr,size,nmemb,(FILE*)stream);
}

bsize_t Bfwrite(const void *ptr, bsize_t size, bsize_t nmemb, BFILE *stream)
{
	return fwrite(ptr,size,nmemb,(FILE*)stream);
}


char *Bstrdup(const char *s)
{
	return strdup(s);
}

char *Bstrcpy(char *dest, const char *src)
{
	return strcpy(dest,src);
}

char *Bstrncpy(char *dest, const char *src, bsize_t n)
{
	return strncpy(dest,src,n);
}

int Bstrcmp(const char *s1, const char *s2)
{
	return strcmp(s1,s2);
}

int Bstrncmp(const char *s1, const char *s2, bsize_t n)
{
	return strncmp(s1,s2,n);
}

int Bstrcasecmp(const char *s1, const char *s2)
{
#ifdef _MSC_VER
	return stricmp(s1,s2);
#else
	return strcasecmp(s1,s2);
#endif
}

int Bstrncasecmp(const char *s1, const char *s2, bsize_t n)
{
#ifdef _MSC_VER
	return strnicmp(s1,s2,n);
#else
	return strncasecmp(s1,s2,n);
#endif
}

char *Bstrcat(char *dest, const char *src)
{
	return strcat(dest,src);
}

char *Bstrncat(char *dest, const char *src, bsize_t n)
{
	return strncat(dest,src,n);
}

bsize_t Bstrlen(const char *s)
{
	return strlen(s);
}

char *Bstrchr(const char *s, int c)
{
	return strchr(s,c);
}

char *Bstrrchr(const char *s, int c)
{
	return strrchr(s,c);
}

int Batoi(const char *nptr)
{
	return atoi(nptr);
}

long Batol(const char *nptr)
{
	return atol(nptr);
}

long int Bstrtol(const char *nptr, char **endptr, int base)
{
	return strtol(nptr,endptr,base);
}

unsigned long int Bstrtoul(const char *nptr, char **endptr, int base)
{
	return strtoul(nptr,endptr,base);
}

void *Bmemcpy(void *dest, const void *src, bsize_t n)
{
	return memcpy(dest,src,n);
}

void *Bmemmove(void *dest, const void *src, bsize_t n)
{
	return memmove(dest,src,n);
}

void *Bmemchr(const void *s, int c, bsize_t n)
{
	return memchr(s,c,n);
}

void *Bmemset(void *s, int c, bsize_t n)
{
	return memset(s,c,n);
}

int Bprintf(const char *format, ...)
{
	va_list ap;
	int r;

	va_start(ap,format);
#ifdef _MSC_VER
	r = _vprintf(format,ap);
#else
	r = vprintf(format,ap);
#endif
	va_end(ap);
	return r;
}

int Bsprintf(char *str, const char *format, ...)
{
	va_list ap;
	int r;

	va_start(ap,format);
#ifdef _MSC_VER
	r = _vsprintf(str,format,ap);
#else
	r = vsprintf(str,format,ap);
#endif
	va_end(ap);
	return r;
}

int Bsnprintf(char *str, bsize_t size, const char *format, ...)
{
	va_list ap;
	int r;

	va_start(ap,format);
#ifdef _MSC_VER
	r = _vsnprintf(str,size,format,ap);
#else
	r = vsnprintf(str,size,format,ap);
#endif
	va_end(ap);
	return r;
}

int Bvsnprintf(char *str, bsize_t size, const char *format, va_list ap)
{
#ifdef _MSC_VER
	return _vsnprintf(str,size,format,ap);
#else
	return vsnprintf(str,size,format,ap);
#endif
}

char *Bgetenv(const char *name)
{
	return getenv(name);
}

char *Bgetcwd(char *buf, bsize_t size)
{
	return getcwd(buf,size);
}

#endif	// __compat_h_macrodef__


//
// Stuff which must be a function
//

int Bgethomedir(char *b, unsigned l)
{
#ifdef WINDOWS
	char *e = getenv("USERPROFILE");
	char cwd[256]=".";
	if (!e) {
		// probably Windows9x
		if (getcwd(cwd,256))
			e=cwd;
	}
#else
	char *e = getenv("HOME");
	if (!e) return -1;
#endif
	if ((strlen(e)+1) > l) return -2;
	strcpy(b,e);
	return strlen(e);
}


int Bcorrectfilename(char *filename, int removefn)
{
#ifdef WINDOWS
	char path[256], *fn;
	int r;

	// Windows uses backslashes so translate all unix-like forwardslashes
	for (fn=path; *fn; fn++) if (*fn == '/') *fn = '\\';

	r = GetFullPathName(filename, 256, path, &fn);
	if (r > 256) return -1;
	if (r == 0) return -1;
	if (removefn && fn) *fn=0;

	strcpy(filename,path);

	return 0;
#else
	char path[256]="", fn[64]="", scratch[256], *ptr, *ptr2, ch;
	char cwd[256], *cwdp = cwd;
	char *tokarr[64];
	int ntok=0, i, j;

	Bgetcwd(cwd, 256);
	Bstrcat(cwd, "/");

	ptr2 = filename;

	ptr = Bstrrchr(ptr2, '/');
	if (!ptr) ptr = Bstrrchr(ptr2, '\\');
	if (ptr) {
		// a path!
		i=0;
		while (ptr2 <= ptr) {
			ch = *(ptr2++);
			if (ch == '\\') ch = '/';
			path[i++] = ch;
		}

		if (path[i-1] != '/') path[i++] = '/';
		path[i] = 0;
	}

	if (!removefn) {
		if (*ptr2 == 0) {
			// a bare path... not good
			return -1;
		}
		Bstrcpy(fn, ptr2);
	}

	// now we have all the bits and pieces, clean it all up
	scratch[0] = 0;

	if (path[0] != '/') {
		// relative path, which means prepend the current dir to the path
		Bstrcat(scratch, cwdp);
	}

	Bstrcat(scratch, path);

	// make sure someone wasn't a smartass by trying dir/../dir
	ptr2 = scratch;
	ptr = ptr2-1;
	do {
		ptr = Bstrchr(ptr+1, '/');
		if (!ptr) break;
		tokarr[ntok++] = ptr;
	} while (1);

	for (i=0; i<ntok; i++) {
		if (Bstrncmp(tokarr[i], "/./", 3) == 0) {
			for (j=i+1; j<ntok; j++) tokarr[j-1] = tokarr[j];
			ntok--;
			i--;
		} else if (Bstrncmp(tokarr[i], "/../", 4) == 0) {
			if (i == 0) {
				for (j=i; j<ntok-1; j++) tokarr[j] = tokarr[j+1];
				ntok--;
				i--;
			} else {
				for (j=i-1; j<ntok-1; j++) tokarr[j] = tokarr[j+2];
				ntok -= 2;
				i -= 2;
			}
		}
	}

	for (i=0; i<ntok; i++) {
		ptr = tokarr[i];
		*(ptr2++) = *(ptr++);
		while ((*ptr != '/') && (*ptr != 0)) *(ptr2++) = *(ptr++);
	}
	*(ptr2) = 0;

	// append the filename
	Bstrcat(scratch, fn);
	Bstrcpy(filename, scratch);

	return 0;
#endif
}


char *Bgetsystemdrives(void)
{
#ifdef WINDOWS
	char *str, *p;
	DWORD drv, mask;
	int number=0;
	
	drv = GetLogicalDrives();
	if (drv == 0) return NULL;

	for (mask=1; mask<0x8000000l; mask<<=1) {
		if ((drv&mask) == 0) continue;
		number++;
	}

	str = p = (char *)malloc(1 + (3*number));
	if (!str) return NULL;

	number = 0;
	for (mask=1; mask<0x8000000l; mask<<=1, number++) {
		if ((drv&mask) == 0) continue;
		*(p++) = 'A' + number;
		*(p++) = ':';
		*(p++) = 0;
	}
	*(p++) = 0;

	return str;
#else
	// Perhaps have Unix OS's put /, /home/user, and /mnt/* in the "drives" list?
	return NULL;
#endif
}


bsize_t Bfilelength(int fd)
{
	struct stat st;
	fstat(fd, &st);
	return(st.st_size);
}


typedef struct {
#ifdef _MSC_VER
	long dir;
	struct _finddata_t fid;
#else
	DIR *dir;
#endif
	struct Bdirent info;
	int status;
	char name[1];
} BDIR_real;

BDIR* Bopendir(const char *name)
{
	BDIR_real *dirr;
#ifdef _MSC_VER
	char *t,*tt;
	t = (char*)malloc(strlen(name)+1+4);
	if (!t) return NULL;
#endif

	dirr = (BDIR_real*)malloc(sizeof(BDIR_real) + strlen(name));
	if (!dirr) {
#ifdef _MSC_VER
		free(t);
#endif
		return NULL;
	}

#ifdef _MSC_VER
	strcpy(t,name);
	tt = t+strlen(name)-1;
	while (*tt == ' ' && tt>t) tt--;
	if (*tt != '/' && *tt != '\\') *(++tt) = '/';
	*(++tt) = '*';
	*(++tt) = '.';
	*(++tt) = '*';
	*(++tt) = 0;
	
	dirr->dir = _findfirst(t,&dirr->fid);
	free(t);
	if (dirr->dir == -1) {
		free(dirr);
		return NULL;
	}
#else
	dirr->dir = opendir(name);
	if (dirr->dir == NULL) {
		free(dirr);
		return NULL;
	}
#endif

	dirr->status = 0;
	strcpy(dirr->name, name);
	
	return (BDIR*)dirr;
}

struct Bdirent*	Breaddir(BDIR *dir)
{
	BDIR_real *dirr = (BDIR_real*)dir;
	struct dirent *de;
	struct stat st;
	char *fn;

#ifdef _MSC_VER
	if (dirr->status > 0) {
		if (_findnext(dirr->dir,&dirr->fid) != 0) {
			dirr->status = -1;
			return NULL;
		}
	}
	dirr->info.namlen = strlen(dirr->fid.name);
	dirr->info.name = dirr->fid.name;
	dirr->status++;
#else
	de = readdir(dirr->dir);
	if (de == NULL) {
		dirr->status = -1;
		return NULL;
	} else {
		dirr->status++;
	}
# ifdef __WATCOMC__
	dirr->info.namlen = strlen(de->d_name);
# else
	dirr->info.namlen = de->d_namlen;
# endif
	dirr->info.name   = de->d_name;
#endif
	dirr->info.mode = 0;
	dirr->info.size = 0;
	dirr->info.mtime = 0;

	fn = malloc(strlen(dirr->name) + 1 + dirr->info.namlen + 1);
	if (fn) {
		sprintf(fn,"%s/%s",dirr->name,dirr->info.name);
		if (!stat(fn, &st)) {
			dirr->info.mode = st.st_mode;
			dirr->info.size = st.st_size;
			dirr->info.mtime = st.st_mtime;
		}
		free(fn);
	}

	return &dirr->info;
}

int Bclosedir(BDIR *dir)
{
	BDIR_real *dirr = (BDIR_real*)dir;
	
#ifdef _MSC_VER
	_findclose(dirr->dir);
#else
	closedir(dirr->dir);
#endif
	free(dirr);

	return 0;
}

