#include <stdio.h>
#include <string.h>

#ifdef WINDOWS
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
//#include <winsock2.h>
#include <winsock.h>
#define ER(x) WSA##x
#define ERS(x) "WSA" x
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#define ER(x) x
#define ERS(x) x
#endif

int ipport = 19014;

#define ADDR_NONE (0)
#define ADDR_IPV4 (1)
#define ADDR_IPV6 (2)
typedef struct {
	unsigned short type;
	union {
		struct sockaddr_in ipv4;
		//struct sockaddr_in6 ipv6;
	} addr;
} multiaddr;



/*
 * Game-land extras
 */

//
// multiaddrtostring() -- Converts the multiaddr address passed into a string
//   which is more human-readable
//
int multiaddrtostring(multiaddr *addr, char *buf, int len)
{
	if (!addr) return -1;
	
	if (addr->type == ADDR_IPV4) {
#ifdef WINDOWS
//		if (!WSAAddressToString((LPSOCKADDR)&addr->addr.ipv4, sizeof(struct sockaddr_in), NULL, buf, (LPDWORD)&len))
//			return 0;
		char *t;
		t = inet_ntoa(addr->addr.ipv4.sin_addr);
		if (t) {
			snprintf(buf,len,"%s:%d",t,ntohs(addr->addr.ipv4.sin_port));
			return 0;
		}
#else
		char t[64];
		int l=64;
	
		if (inet_ntop(AF_INET, &addr->addr.ipv4, t, l)) {
			snprintf(buf,len,"%s:%d",t,ntohs(addr->addr.ipv4.sin_port));
			return 0;
		}
#endif
	} else {
		strncpy(buf,"Invalid",len);
	}

	return -1;
}


//
// multistringtoaddr() -- Resolves a hostname and stores the result in a multiaddr
//
static int multistringtoaddr(char *str, multiaddr *addr)
{
	struct hostent *he;
	int port = ipport, t;
	char *host, *p, *p2;

	host = alloca(strlen(str)+1);
	strcpy(host,str);

	p = strchr(host,':');
	if (p) {
		*(p++) = 0;
		t = strtol(p, &p2, 10);
		if (p != p2 && *p2 == 0) port = t;
	}

	memset(addr, 0, sizeof(multiaddr));
	he = gethostbyname(host);
	if (!he) {
		switch (h_errno) {
			case HOST_NOT_FOUND: printf("Name resolution error: %s: Host not found\n", host); break;
			case NO_DATA: printf("Name resolution error: %s: No address for this hostname\n", host); break;
			case NO_RECOVERY: printf("Name resolution error: %s: Unrecoverable error\n", host); break;
			case TRY_AGAIN: printf("Name resolution error: %s: Temporary name server error\n", host); break;
		}
		
		return -1;
	}

	if (he->h_addrtype == AF_INET) {
		addr->type = ADDR_IPV4;
		addr->addr.ipv4.sin_family = AF_INET;
		memcpy(&addr->addr.ipv4.sin_addr, he->h_addr_list[0], he->h_length);
		addr->addr.ipv4.sin_port = htons(port);
	}

	return 0;
}


//
// multicompareaddr() -- Compares two addresses. Returns nonzero on difference.
//
int multicompareaddr(multiaddr *addr1, multiaddr *addr2)
{
	if (addr1->type != addr2->type) return 1;
	switch (addr1->type) {
		case ADDR_NONE:
			return 0;
		case ADDR_IPV4:
			if (addr1->addr.ipv4.sin_addr.s_addr != addr2->addr.ipv4.sin_addr.s_addr)
				return 1;
			if (addr1->addr.ipv4.sin_port != addr2->addr.ipv4.sin_port)
				return 1;
			return 0;
		default:
			return 1;
	}
}



int main(int argc, char **argv)
{
	char type;
	long clock;
	multiaddr addr;
	int i,j;
	char workbuf[64];
	FILE *fp;
	
	if (argc<2) {
		puts("grokmultidump <dumpfile>");
		return -1;
	}

	fp = fopen(argv[1], "r");
	if (!fp) {
		printf("Failed opening %s\n", argv[1]);
		return -1;
	}

	while (!feof(fp)) {
		type = fgetc(fp);
		clock = (long)fgetc(fp) | ((long)fgetc(fp) << 8) | ((long)fgetc(fp) << 16) | ((long)fgetc(fp) << 24);

		switch (type) {
			case 0:
				printf("INIT %10d\n", clock);
				break;
			case 1:
				i = fgetc(fp);
				fread(workbuf, i, 1, fp);
				workbuf[i] = 0;
				printf("ADDP %10d %s\n", clock, workbuf);
				break;
			case 2:
				i = fgetc(fp);
				memset(&addr, 0, sizeof(multiaddr));
				fread(&addr, i, 1, fp);
				multiaddrtostring(&addr, workbuf, 64);
				printf("SEND %010d %-24s", clock, workbuf);
				j = (long)fgetc(fp) | ((long)fgetc(fp) << 8);
				i = 0;
				while (j) {
					printf(" %02x", fgetc(fp));
					if (--j > 0) {
						i++;
						if (i == 16) {
							printf("\n                                        ");
							i = 0;
						} else if (i == 8) { printf(" "); }
					}
				}
				printf("\n");
				break;
			case 3:
				i = fgetc(fp);
				memset(&addr, 0, sizeof(multiaddr));
				fread(&addr, i, 1, fp);
				multiaddrtostring(&addr, workbuf, 64);
				printf("RECV %010d %-24s", clock, workbuf);
				j = (long)fgetc(fp) | ((long)fgetc(fp) << 8);
				i = 0;
				while (j) {
					printf(" %02x", fgetc(fp));
					if (--j > 0) {
						i++;
						if (i == 16) {
							printf("\n                                        ");
							i = 0;
						} else if (i == 8) { printf(" "); }
					}
				}
				printf("\n");
				break;
			case 4:
				printf("CLOS %10d\n", clock);
				break;
		}
	}

	fclose(fp);
	
	return 0;
}
